unit wbfrm;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, OleCtrls, SHDocVw_EWB, EwbCore, EmbeddedWB, dx_scene, dx_vgcore,
  dx_vglayer, ActiveX, MSHTML_EWB;

type
  TfrmWebBrowser = class(TForm)
    dxScene1: TdxScene;
    Root1: TdxVisualObject;
    WBBuffer: TdxBufferLayer;
    EmbeddedWB1: TEmbeddedWB;
    ToolLayer: TdxGUIObjectLayer;
    Root2: TvxLayout;
    ToolBar1: TvxToolBar;
    textUrl: TvxTextBox;
    btnNav: TvxCornerButton;
    StatusLayer: TdxGUIObjectLayer;
    Root3: TvxLayout;
    StatusBar1: TvxStatusBar;
    TrackBar1: TvxTrackBar;
    Label1: TvxLabel;
    Label2: TvxLabel;
    labelStatus: TvxLabel;
    btnBack: TvxCornerButton;
    VScrollLayer: TdxGUIObjectLayer;
    Root4: TvxLayout;
    VScroll: TvxScrollBar;
    procedure btnNavClick(Sender: TObject);
    procedure EmbeddedWB1DocumentComplete(ASender: TObject;
      const pDisp: IDispatch; var URL: OleVariant);
    procedure WBBufferLayerMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure WBBufferLayerMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure WBBufferLayerMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure EmbeddedWB1UpdateUI(Sender: TObject; var Rezult: HRESULT);
    procedure textUrlChange(Sender: TObject);
    procedure WBBufferMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TdxPoint; var Handled: Boolean);
    procedure TrackBar1Change(Sender: TObject);
    procedure EmbeddedWB1StatusTextChange(ASender: TObject;
      const Text: WideString);
    procedure btnBackClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure EmbeddedWB1NewWindow2(ASender: TObject;
      var ppDisp: IDispatch; var Cancel: WordBool);
    procedure FormDestroy(Sender: TObject);
    procedure VScrollChange(Sender: TObject);
  private
    GDIBuffer: TBitmap;
    procedure UpdateBuffer;
    procedure SettingScroll;
    { Private declarations }
  public
    { Public declarations }
  end;

var
  frmWebBrowser: TfrmWebBrowser;

implementation

{$R *.dfm}

function KeysFromShift(Shift: TShiftState): cardinal;
begin
  Result := 0;
  if ssCtrl in Shift then
    Result := Result or MK_CONTROL;
  if ssShift in Shift then
    Result := Result or MK_SHIFT;
  if ssAlt in Shift then
    Result := Result or MK_ALT;
  if ssLeft in Shift then
    Result := Result or MK_LBUTTON;
  if ssRight in Shift then
    Result := Result or MK_RBUTTON;
end;

function GetBrowserElement(Browser: TEmbeddedWB; out Element: IHTMLElement2): HResult;
var
  pDoc:          IHTMLDocument2;
  pElemParent:   IHTMLElement;
  pElement:      IHTMLElement;
  pElement2:     IHTMLElement2;
begin

  // Check browser document
  if Assigned(Browser.Document) then
  begin
     // QI for document 2 interface
     result:=Browser.Document.QueryInterface(IHTMLDocument2, pDoc);
     // Check result
     if Succeeded(result) then
     begin
        // Resource protection
        try
           // Check document body
           if Assigned(pDoc.Body) then
           begin
              // QI for body element
              result:=pDoc.Body.QueryInterface(IHTMLElement, pElement);
              // Check result
              if Succeeded(result) then
              begin
                 // Resource protection
                 try
                    // Get parent element
                    pElemParent:=pElement.parentElement;
                    // Check parent
                    if Assigned(pElemParent) then
                    begin
                       // Resource protection
                       try
                          // Get the element 2 interface
                          if Succeeded(pElemParent.QueryInterface(IHTMLElement2, pElement2)) then
                          begin
                             // Check parent interface for zero client width or height
                             if (pElement2.clientWidth = 0) or (pElement2.clientHeight = 0) then
                                // QI for element 2 interface
                                result:=pElement.QueryInterface(IHTMLElement2, Element)
                             else
                                // We have the element we are after
                                Element:=pElement2;
                          end
                          else
                             // QI for element 2 interface
                             result:=pElement.QueryInterface(IHTMLElement2, Element);
                       finally
                          // Release the interface
                          pElemParent:=nil;
                       end;
                    end
                    else
                       // Return nearest COM error
                       result:=E_NOINTERFACE;
                 finally
                    // Release the interface
                    pElement:=nil;
                 end;
              end;
           end
           else
              // Return nearest COM error
              result:=E_NOINTERFACE;
        finally
           // Release the interface
           pDoc:=nil;
        end;
     end;
  end
  else
     // Return nearest COM error
     result:=E_NOINTERFACE;

end;

procedure TfrmWebBrowser.FormCreate(Sender: TObject);
begin
  // Create GDI Buffer - because IE only can draw to DC
  GDIBuffer := TBitmap.Create;
  GDIBuffer.PixelFormat := pf32bit;
  // disable control update
  SendMessage(EmbeddedWB1.Handle, WM_SETREDRAW, 0, 0);
  // update scroll
  SettingScroll;
end;

procedure TfrmWebBrowser.FormDestroy(Sender: TObject);
begin
  GDIBuffer.Free;
end;

procedure TfrmWebBrowser.textUrlChange(Sender: TObject);
begin
  btnNavClick(Sender);
end;

procedure TfrmWebBrowser.btnNavClick(Sender: TObject);
begin
  EmbeddedWB1.Navigate(textUrl.text);
end;

procedure TfrmWebBrowser.UpdateBuffer;
var
  View: IViewObject2;
  Doc: IHTMLDocument2;
  DS: IDisplayServices;
  C: IHTMLCaret;
  r: TRect;
  V: integer;
  P: TPoint;
begin
  (* Get WB Document *)
  Doc := IHTMLDocument2(EmbeddedWB1.Document);
  if not Assigned(Doc) then Exit;
  { Set size to EmbeddedWB control }
  EmbeddedWB1.Width := WBBuffer.LayerWidth;
  EmbeddedWB1.Height := WBBuffer.LayerHeight;
  (* Make screenshot *)
  view := EmbeddedWB1.controlinterface as IViewObject2;
  r := rect(0, 0, WBBuffer.LayerWidth, WBBuffer.LayerHeight);
  (* Resize buffer (if need) *)
  GDIBuffer.Width := WBBuffer.LayerWidth;
  GDIBuffer.Height := WBBuffer.LayerHeight;
  (* Draw to GDI buffer *)
  view.Draw(DVASPECT_CONTENT, 1, nil, nil, 0, GDIBuffer.canvas.Handle, @r, nil, nil, 0);
  (* Draw Caret *)
  Doc.QueryInterface(IDisplayServices, DS);
  if Assigned(DS) then
  begin
    DS.GetCaret(C);
    if Assigned(C) then
    begin
      C.IsVisible(V);
      if V > 0 then
      begin
        C.GetLocation(P, 1);
        GDIBuffer.Canvas.Brush.Color := clBlack;
        GDIBuffer.Canvas.FillRect(Rect(P.x, P.y, P.x + 2, P.y + 18));
      end;
    end;
  end;
  (* Copy to WBBuffer *)
  WBBuffer.Buffer.Assign(GDIBuffer);
  (* Fill A-channel - because IE not set it *)
  WBBuffer.Buffer.FillAlpha($FF);
  (* Force Update *)
  WBBuffer.Repaint;
end;

procedure TfrmWebBrowser.EmbeddedWB1DocumentComplete(ASender: TObject;
  const pDisp: IDispatch; var URL: OleVariant);
begin
  UpdateBuffer;
  SettingScroll;
end;

procedure TfrmWebBrowser.WBBufferLayerMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Msg: TWMMouse;
  R: LRESULT;
begin
  case Button of
    mbLeft: Msg.Msg := WM_LBUTTONDOWN;
    mbRight: Msg.Msg := WM_RBUTTONDOWN;
  end;
  Msg.Keys := KeysFromShift(Shift);
  Msg.XPos := round(X);
  Msg.YPos := round(Y);
  with TMessage(Msg) do
    SendMessage(EmbeddedWB1.HWND, Msg, wParam, lParam);
  // need to force capture by IE
  SetCapture(dxScene1.Handle);
end;

procedure TfrmWebBrowser.WBBufferLayerMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
  Msg: TWMMouse;
  R: LRESULT;
begin
  Msg.Msg := WM_MOUSEMOVE;
  Msg.Keys := KeysFromShift(Shift);
  Msg.XPos := round(X);
  Msg.YPos := round(Y);
  with TMessage(Msg) do
    SendMessage(EmbeddedWB1.HWND, Msg, wParam, lParam);
end;

procedure TfrmWebBrowser.WBBufferLayerMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Msg: TWMMouse;
  R: LRESULT;
begin
  case Button of
    mbLeft: Msg.Msg := WM_LBUTTONUP;
    mbRight: Msg.Msg := WM_RBUTTONUP;
  end;
  Msg.Keys := KeysFromShift(Shift);
  Msg.XPos := round(X);
  Msg.YPos := round(Y);
  with TMessage(Msg) do
    SendMessage(EmbeddedWB1.HWND, Msg, wParam, lParam);
end;

procedure TfrmWebBrowser.EmbeddedWB1UpdateUI(Sender: TObject;
  var Rezult: HRESULT);
begin
  // update buffer
  UpdateBuffer;
end;

procedure TfrmWebBrowser.TrackBar1Change(Sender: TObject);
begin
  WBBuffer.RotateAngle.Z := TrackBar1.Value;
end;

procedure TfrmWebBrowser.EmbeddedWB1StatusTextChange(ASender: TObject;
  const Text: WideString);
begin
  labelStatus.Text := Text;
  // update buffer
  UpdateBuffer;
end;

procedure TfrmWebBrowser.btnBackClick(Sender: TObject);
begin
  EmbeddedWB1.GoBack;
end;

procedure TfrmWebBrowser.EmbeddedWB1NewWindow2(ASender: TObject;
  var ppDisp: IDispatch; var Cancel: WordBool);
begin
  Cancel := true;
end;

procedure TfrmWebBrowser.SettingScroll;
var
  pElement: IHTMLElement2;
  H: integer;
begin
  // Get the browser body element interface
  if Succeeded(GetBrowserElement(EmbeddedWB1, pElement)) then
  begin
    try
      H := pElement.ScrollHeight;
    finally
      pElement:=nil;
    end;
  end
  else
    H := 0;
  if H > EmbeddedWB1.Height then
  begin
    VScroll.Visible := true;
    VScroll.Max := H;
    VScroll.Value := EmbeddedWB1.VScrollBarPosition;
    VScroll.ViewportSize := EmbeddedWB1.Height;
  end
  else
    VScroll.Visible := false;
end;

procedure TfrmWebBrowser.VScrollChange(Sender: TObject);
begin
  EmbeddedWB1.ScrollToPosition(0, round(VScroll.Value));
  UpdateBuffer;
end;

procedure TfrmWebBrowser.WBBufferMouseWheel(Sender: TObject;
  Shift: TShiftState; WheelDelta: Integer; MousePos: TdxPoint;
  var Handled: Boolean);
begin
  EmbeddedWB1.ScrollToPosition(0, EmbeddedWB1.VScrollBarPosition - WheelDelta);
  SettingScroll;
  Handled := true;
end;

end.
